'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function cloneKey(args) {
    var length = args.length;
    var key = Array(length);
    for (var index = 0; index < length; ++index) {
        key[index] = args[index];
    }
    return key;
}
function getDefault(type, value, defaultValue) {
    return typeof value === type ? value : defaultValue;
}
function getEntry(node) {
    return [node.k, node.v];
}
function isMemoized(fn) {
    return typeof fn === 'function' && fn.isMemoized;
}
function isSameValueZero(a, b) {
    return a === b || (a !== a && b !== b);
}

var Cache = /** @class */ (function () {
    function Cache(options) {
        this.size = 0;
        this.h = null;
        this.oa = null;
        this.od = null;
        this.oh = null;
        this.ou = null;
        this.t = null;
        var transformKey = getDefault('function', options.transformKey);
        this.a = getDefault('function', options.isArgEqual, isSameValueZero);
        this.l = getDefault('number', options.maxSize, 1);
        this.m = getDefault('function', options.isKeyEqual, this.e);
        this.p = getDefault('boolean', options.async, false);
        if (transformKey || options.isKeyEqual === this.m) {
            this.k = transformKey
                ? function (args) { return transformKey(cloneKey(args)); }
                : cloneKey;
        }
    }
    Cache.prototype.clear = function () {
        this.h = this.t = null;
        this.size = 0;
    };
    Cache.prototype.delete = function (key) {
        var node = this.gt(key);
        if (node) {
            this.d(node);
            this.od && this.od.n(node);
            return true;
        }
        return false;
    };
    Cache.prototype.entries = function () {
        var entries = [];
        var node = this.h;
        while (node != null) {
            entries.push(getEntry(node));
            node = node.n;
        }
        return entries;
    };
    Cache.prototype.get = function (key) {
        var node = this.gt(key);
        if (node) {
            node !== this.h && this.u(node);
            return node.v;
        }
    };
    Cache.prototype.has = function (key) {
        return !!this.gt(key);
    };
    Cache.prototype.off = function (type, listener) {
        var emitter = this.og(type);
        if (emitter) {
            emitter.r(listener);
            !emitter.s && this.os(type, null);
        }
    };
    Cache.prototype.on = function (type, listener) {
        var emitter = this.og(type);
        if (!emitter) {
            emitter = this.c(type);
            this.os(type, emitter);
        }
        emitter.a(listener);
        return listener;
    };
    Cache.prototype.set = function (key, value) {
        var normalizedKey = this.k ? this.k(key) : key;
        var node = this.g(normalizedKey);
        if (node) {
            node.v = value;
            node !== this.h && this.u(node);
            this.ou && this.ou.n(node);
        }
        else {
            node = this.n(normalizedKey, value);
            this.oa && this.oa.n(node);
        }
        return node;
    };
    Cache.prototype.c = function (type) {
        var listeners = [];
        // eslint-disable-next-line @typescript-eslint/no-this-alias
        var cache = this;
        var size = 0;
        function a(listener) {
            if (listeners.indexOf(listener) === -1) {
                listeners.push(listener);
                ++size;
            }
        }
        function n(node, reason) {
            var entry = getEntry(node);
            for (var index = 0; index < size; ++index) {
                listeners[index]({
                    cache: cache,
                    entry: entry,
                    reason: reason,
                    type: type,
                });
            }
        }
        function r(listener) {
            var index = listeners.indexOf(listener);
            if (index !== -1) {
                listeners.splice(index, 1);
                --size;
            }
        }
        return {
            get s() {
                return size;
            },
            a: a,
            n: n,
            r: r,
        };
    };
    Cache.prototype.d = function (node) {
        var next = node.n;
        var prev = node.p;
        if (next) {
            next.p = prev;
        }
        else {
            this.t = prev;
        }
        if (prev) {
            prev.n = next;
        }
        else {
            this.h = next;
        }
        --this.size;
    };
    Cache.prototype.e = function (prevKey, nextKey) {
        var length = nextKey.length;
        if (prevKey.length !== length) {
            return false;
        }
        if (length === 1) {
            return this.a(prevKey[0], nextKey[0]);
        }
        for (var index = 0; index < length; ++index) {
            if (!this.a(prevKey[index], nextKey[index])) {
                return false;
            }
        }
        return true;
    };
    Cache.prototype.g = function (key) {
        var node = this.h;
        if (!node) {
            return;
        }
        if (this.m(node.k, key)) {
            return node;
        }
        if (this.h === this.t) {
            return;
        }
        node = node.n;
        while (node) {
            if (this.m(node.k, key)) {
                return node;
            }
            node = node.n;
        }
    };
    Cache.prototype.gt = function (key) {
        return this.g(this.k ? this.k(key) : key);
    };
    Cache.prototype.n = function (key, value) {
        var _this = this;
        var prevHead = this.h;
        var prevTail = this.t;
        var node = { k: key, n: prevHead, p: null, v: value };
        if (this.p) {
            node.v = value.then(function (value) {
                _this.ou && _this.g(key) && _this.ou.n(node, 'resolved');
                return value;
            }, function (error) {
                if (_this.g(key)) {
                    _this.d(node);
                    _this.od && _this.od.n(node, 'rejected');
                }
                throw error;
            });
        }
        this.h = node;
        if (prevHead) {
            prevHead.p = node;
        }
        else {
            this.t = node;
        }
        if (++this.size > this.l && prevTail) {
            this.d(prevTail);
            this.od && this.od.n(prevTail, 'evicted');
        }
        return node;
    };
    Cache.prototype.og = function (type) {
        if (type === 'add') {
            return this.oa;
        }
        if (type === 'delete') {
            return this.od;
        }
        if (type === 'hit') {
            return this.oh;
        }
        if (type === 'update') {
            return this.ou;
        }
        return null;
    };
    Cache.prototype.os = function (type, emitter) {
        if (type === 'add') {
            this.oa = emitter;
        }
        else if (type === 'delete') {
            this.od = emitter;
        }
        else if (type === 'hit') {
            this.oh = emitter;
        }
        else if (type === 'update') {
            this.ou = emitter;
        }
    };
    Cache.prototype.u = function (node) {
        var next = node.n;
        var prev = node.p;
        if (next) {
            next.p = prev;
        }
        if (prev) {
            prev.n = next;
        }
        this.h.p = node;
        node.n = this.h;
        node.p = null;
        this.h = node;
        if (node === this.t) {
            this.t = prev;
        }
    };
    return Cache;
}());

var memoize = function memoize(fn, passedOptions) {
    if (passedOptions === void 0) { passedOptions = {}; }
    if (typeof fn !== 'function') {
        throw new TypeError("Expected first parameter to be function; received ".concat(typeof fn));
    }
    if (isMemoized(fn)) {
        return memoize(fn.fn, Object.assign({}, fn.options, passedOptions));
    }
    var memoized = function memoized() {
        var cache = memoized.cache;
        var transformKey = cache.k;
        var key = transformKey
            ? transformKey(arguments)
            : arguments;
        var node = cache.g(key);
        if (node === cache.h) {
            cache.oh && cache.oh.n(node);
        }
        else if (node) {
            cache.u(node);
            cache.oh && cache.oh.n(node);
            cache.ou && cache.ou.n(node);
        }
        else {
            node = cache.n(transformKey ? key : cloneKey(key), 
            // @ts-expect-error - allow usage of arguments as pass-through to fn
            fn.apply(this, arguments));
            cache.oa && cache.oa.n(node);
        }
        return node.v;
    };
    memoized.cache = new Cache(passedOptions);
    memoized.fn = fn;
    memoized.isMemoized = true;
    memoized.options = passedOptions;
    return memoized;
};

exports.Cache = Cache;
exports.default = memoize;
